! ******************************************************* !
!
subroutine calc_renormalized_pk_2loop
!                                                         !
!                                                         !
!             Time-stamp: <2011-03-02 20:02:33 t-okamura> !
! ******************************************************* !
!
!     Computing renormalized power spectra for 2-loop order 
!     (or higher-order Born approximation) using hybrid integration 
!     scheme (Gaussian quadrature for time integral (2D), Monte Calro 
!     method for Fourier integral (4D))
!
!     This code should be compiled with CUBA library, libcuba.a
!
  use data, only : pi, itype, eeta, ikmax, ak, pk, ik_max, isize, irank, &
       & ik_init, k_cut_min, k_cut_max, pk_2loop11, pk_2loop12, pk_2loop22
  implicit none
 
  integer, parameter :: ndim4 = 4
  integer, parameter :: ncomp = 3
  double precision, parameter :: epsrel = 1d-3
  double precision, parameter :: epsabs = 1d-12
  integer, parameter :: verbose = 0
  integer, parameter :: last = 4
  integer, parameter :: mineval = 0
  integer, parameter :: seed = 0
!     for Suave 
  integer, parameter :: maxeval = 50000
!     for Cuhre
!  integer, parameter :: maxeval = 100000
  integer, parameter ::  nnew = 1000
  double precision, parameter :: flatness = 15.d0
  integer, parameter :: key = 0
  double precision integral(ncomp), error(ncomp), prob(ncomp), userdata(2)
  integer nregions, neval, fail
! ---------------------------------------------------
  integer ik
  double precision k, fact, kmin, kmax
  external integ_pk_2loop
! -------------------------------------------------

!
! //////// calculation of 2-loop P(k) //////// !
!
  kmin = ak(1)
  kmax = ak(ik_max)

  pk_2loop11(:) = 0.d0
  pk_2loop12(:) = 0.d0
  pk_2loop22(:) = 0.d0

  !$OMP PARALLEL DEFAULT(SHARED), &
  !$OMP & PRIVATE(ik, k, userdata, fact, nregions, &
  !$OMP &        neval, fail, integral, error, prob )
  !$OMP DO SCHEDULE(DYNAMIC,1)
  do ik = ik_init + irank, ik_max, isize
     k = ak(ik)

!     skip the integration for small-k or large-k mode

     if( (k < k_cut_min) .or. (k > k_cut_max) ) cycle

!     integration by Monte Carlo method

     userdata(1) = k
     userdata(2) = ( log(kmax/k+1.d0) - log(kmin/kmax) ) &
          & * log(kmax/kmin) * 8.d0 

     call suave(ndim4, ncomp, integ_pk_2loop, userdata,     &
          &     epsrel, epsabs, verbose + last, seed,       &
          &     mineval, maxeval,                           &
          &     nnew, flatness,                             &
          &     nregions, neval, fail, integral, error, prob)
!!!         call cuhre(ndim4, ncomp, integ_pk_2loop, userdata,           &
!!!              &     epsrel, epsabs, verbose + last, mineval, maxeval, &
!!!             &     key,                                              &
!!!              &     nregions, neval, fail, integral, error, prob)

     fact = k**6 / (2.d0*pi)**4 * 4.d0
     pk_2loop11(ik) = integral(1) * fact
     pk_2loop12(ik) = integral(2) * fact
     pk_2loop22(ik) = integral(3) * fact

     if(itype == 0) then 
        pk_2loop11(ik) = 16.d0 * pk_2loop11(ik)
        pk_2loop12(ik) = 16.d0 * pk_2loop12(ik)
        pk_2loop22(ik) = 16.d0 * pk_2loop22(ik)
     elseif(itype == 1) then
        pk_2loop11(ik) = 8.d0 * pk_2loop11(ik)
        pk_2loop12(ik) = 8.d0 * pk_2loop12(ik)
        pk_2loop22(ik) = 8.d0 * pk_2loop22(ik)
     endif
  
  enddo
  !$OMP END DO
  !$OMP END PARALLEL
  
  return
end subroutine calc_renormalized_pk_2loop
!
! ******************************************************* !
!
subroutine integ_pk_2loop(ndim, xx, ncomp, ff, userdata)
!
! ******************************************************* !
!
!     integrand of 2-loop correction
!
  use data,only : kmin, kmax, eta0, eta, eeta0, itype
  implicit none

  integer, intent(in) :: ndim, ncomp
  double precision, intent(in) ::  xx(ndim), userdata(2)
  double precision, intent(out) :: ff(ncomp)
! --------------------------------------- !
  double precision integ_pk_A11, integ_pk_A12, integ_pk_A22
  double precision integ_pk_B11, integ_pk_B12, integ_pk_B22
  double precision k, x, mu, y, nu, jacobian, fact, dlnk
  double precision xabs, yabs, xabs2, yabs2
  double precision J1, J2, K1, K2
  double precision kp, q, pq, p, kpq
  double precision pklin_p, pklin_q, pklin_kp, pklin_pq, pklin_kpq

  k = userdata(1)
  jacobian = userdata(2)
  dlnk = log(kmax/kmin)
  
  x = exp( -dlnk + ( log(kmax/k+1.d0) + dlnk ) * xx(1) )
  mu = -1.d0 + 2.d0 * xx(2)
  y = exp( dlnk * (-1.d0 + 2.d0 * xx(3) ) )
  nu = -1.d0 + 2.d0 * xx(4)

  xabs2 = 1.d0 + x*x -2.d0*x*mu
  yabs2 = 1.d0 + y*y -2.d0*y*nu

  if( (xabs2 >= 0.d0) .and. (yabs2  >= 0.d0) & 
       & .and. (x < 0.5d0 .or. (x >= 0.5d0 .and. (mu < 0.5d0/x))) &
       & .and. (y < 0.5d0 .or. (y >= 0.5d0 .and. (nu < 0.5d0/y))) ) then 
     
     xabs = sqrt(xabs2)
     yabs = sqrt(yabs2)

!
! ///// A-part ///// !
!
     kp = k * xabs
     q =  k * x * y
     pq = k * x * yabs 

     if( ((kp >= kmin) .and. (kp <= kmax)) &
          & .and. ((q >= kmin) .and. (q <= kmax)) &
          & .and. ((pq >= kmin) .and. (pq <= kmax)) ) then 
        
        call get_func_J(itype, k, x, y, mu, nu, xabs2, yabs2, &
             & eta, eta0, J1, J2)
        call find_pk(q, pklin_q)
        call find_pk(kp, pklin_kp)
        call find_pk(pq, pklin_pq)
        
        fact = x**3 * pklin_q * pklin_kp * pklin_pq
        
        integ_pk_A11 = fact * J1 * J1
        integ_pk_A12 = fact * J1 * J2
        integ_pk_A22 = fact * J2 * J2
     else
        integ_pk_A11 = 0.d0
        integ_pk_A12 = 0.d0
        integ_pk_A22 = 0.d0
     endif
!
! ///// B-part ///// !
!
     p = k * x
     q = k * y * xabs  
     kpq = k * xabs * yabs
     
     if( ((p >= kmin) .and. (p <= kmax))  &
          & .and. ((q >= kmin) .and. (q <= kmax)) &
          & .and. ((kpq >= kmin) .and. (kpq <= kmax)) ) then 

        call get_func_K(itype, k, x, y, mu, nu, xabs2, yabs2, &
             & eta, eta0, K1, K2)  
        call find_pk(q, pklin_q)
        call find_pk(p, pklin_p)
        call find_pk(kpq, pklin_kpq)
        
        fact = xabs**3 * pklin_q * pklin_p * pklin_kpq
        
        integ_pk_B11 = fact * K1 * K1
        integ_pk_B12 = fact * K1 * K2
        integ_pk_B22 = fact * K2 * K2
        
     else
        integ_pk_B11 = 0.d0
        integ_pk_B12 = 0.d0
        integ_pk_B22 = 0.d0
     endif
  else
     integ_pk_A11 = 0.0d0
     integ_pk_A12 = 0.0d0
     integ_pk_A22 = 0.0d0
     integ_pk_B11 = 0.0d0
     integ_pk_B12 = 0.0d0
     integ_pk_B22 = 0.0d0
  endif

  fact = (x * y)**3 * 0.5d0 * jacobian * eeta0**6
  ff(1) = (integ_pk_A11 + integ_pk_B11) * fact
  ff(2) = (integ_pk_A12 + integ_pk_B12) * fact
  ff(3) = (integ_pk_A22 + integ_pk_B22) * fact

  return
end subroutine integ_pk_2loop
!
! ******************************************************* !
!
subroutine get_func_J(itype, k, x, y, mu, nu, xabs2, yabs2, &
     & eta, eta0, func_J1, func_J2)
!
! ******************************************************* !
!
!     computing function J_1 & J_2 using method of Gaussian quadratures 
!
  implicit none
  integer, parameter :: ieta_max = 40
  integer, intent(in) :: itype
  double precision, intent(in) :: k, x, y, mu, nu, xabs2, yabs2, eta, eta0
  double precision, intent(out) :: func_J1, func_J2
  integer ieta
  double precision int_J1, int_J2, func_J1a
  double precision k1, k2, k3, k4, k5, ss
  double precision ff, gg, hh, ii, jj
  double precision weta(ieta_max), xeta(ieta_max)
  double precision fact, xy
  double precision ef, eg, eh, ei, ej
! ------------------------------------------

  k1 = k
  k2 = k * x
  k3 = k * sqrt(xabs2)
  k4 = k2 * y
  k5 = k2 * sqrt(yabs2)

  if(max(k1,k2,k3,k4,k5) <= 5d-3) then 
!
!     using the linear propagator, integration is analytically performed
!
     ss = eta - eta0
     xy = x * y
     fact = xabs2 * yabs2 * xy

     eh = exp(ss)
     eg = eh * eh         ! exp(2ss)    
     ef = eg * eh         ! exp(3ss)
     ei = 1.d0 / sqrt(eh) ! exp(-0.5ss)
     ej = ei / eh         ! exp(-1.5ss)
     
     ! //// function J1 //// !
     ff = 28.d0*x*x*y*mu * ( nu*nu - 1.d0 )  &
          & - 9.d0 * mu * ( y - 7.d0*nu + 6.d0*y*nu*nu ) &
          & + 7.d0 *x* (5.d0 -14.d0*mu*mu ) * nu &
          & + x*y * (23.d0 -58.d0*nu*nu &
          & + 14.d0*mu*mu * (1.d0 + 6.d0*nu*nu ) )
     ff = ff / 504.d0
     gg = ( -7.d0*mu + x*(-3.d0 + 10.d0*mu*mu ) ) &
          & * ( -5.d0*nu + y*(-1.d0 + 6.d0*nu*nu ) )
     gg = -gg / 140.d0
     hh = -4.d0*x*x*y*mu * ( nu*nu -1.d0 ) &
          & + 5.d0*mu * ( y + nu -2.d0*y*nu*nu ) &
          & + x * ( 1.d0 - 6.d0*mu*mu ) * nu &
          & + x*y * (-3.d0 + 2.d0*nu*nu &
          & + 6.d0*mu*mu * ( 2.d0*nu*nu -1.d0 ) )
     hh = hh / 40.d0
     ii = -2.d0/35.d0 * ( x - mu ) * ( nu*nu - 1.d0 ) * y
     jj = -2.d0*( mu*mu - 1.d0 )*nu - y*( nu*nu + 1.d0 )  &
          & - y*mu*( 7.d0*x - 8.d0*mu + (-7.d0*x + 6.d0*mu)*nu*nu ) 
     jj = 2.d0/315.d0 * jj * x

     func_J1 = ff * ef + gg * eg + hh * eh + ii * ei + jj * ej
     func_J1 = func_J1 / fact

     ! //// function J2 ///// !     
     ff = 4.d0*x*x*y*mu * ( nu*nu - 1.d0 )  &
          & - 3.d0 * mu * ( y - 7.d0*nu + 6.d0*y*nu*nu ) &
          & - 7.d0 *x* ( 1.d0 + 2.d0*mu*mu ) * nu &
          & + x*y * ( 5.d0 + 2.d0*nu*nu  &
          & + 2.d0*mu*mu * ( 1.d0 + 6.d0*nu*nu ) )
     ff = ff / 168.d0
     gg = ( -7.d0*mu + x*( 1.d0 + 6.d0*mu*mu ) )     &
     &        * ( -5.d0*nu + y*(-1.d0 + 6.d0*nu*nu ) ) 
     gg = -gg / 140.d0
     hh = -4.d0*x*x*y*mu * ( nu*nu - 1.d0 )         &
          & + 5.d0*mu * ( y + nu - 2.d0*y*nu*nu ) &
          & + x * ( 1.d0 - 6.d0*mu*mu ) * nu      &
          & + x*y * ( -3.d0 + 2.d0*nu*nu          &
          & + 6.d0*mu*mu * ( 2.d0*nu*nu - 1.d0 ) )
     hh = hh / 40.d0
     ii = x * ( 2.d0*mu + x*(-1.d0+5.d0*x*mu-6.d0*mu*mu) )  &
          & * ( nu*nu -1.d0 ) / 35.d0
     jj = -jj * 1.5d0

     func_J2 = ff * ef + gg * eg + hh * eh + ii * ei + jj * ej
     func_J2 = func_J2 / fact

  elseif(max(k1,k2,k3,k4,k5) > 5d-3) then 
!     numerical integration by Gaussian quadratures
     func_J1 = 0.d0
     func_J2 = 0.d0
     call gauleg(eta0, eta, xeta, weta, ieta_max)

     do ieta = 1, ieta_max
        call int_J(k, x, y, mu, nu, xabs2, yabs2, eta, eta0, xeta(ieta), &
             & itype, int_J1, int_J2)
        func_J1 = func_J1 + weta(ieta) * int_J1
        func_J2 = func_J2 + weta(ieta) * int_J2
     enddo

  endif
  return
end subroutine get_func_J
! ******************************************************* !
!
subroutine int_J(k, x, y, mu, nu, xabs2, yabs2, eta, eta0, eta1, &
     & itype, int_J1, int_J2)
!
! ******************************************************* !

  implicit none
  integer, intent(in) :: itype
  double precision, intent(in) :: k, x, y, mu, nu, xabs2, yabs2, eta, eta0, eta1
  double precision, intent(out) :: int_J1, int_J2
  double precision kxmu
  double precision G11a, G12a, G21a, G22a, G11c, G12c, G21c, G22c
  double precision I1, I2, A, B
  double precision gamma112, gamma121, gamma222
! ---------------------------------

  kxmu = k * sqrt(xabs2)

  if(itype == 0) then 
     call G_ab_RPT(k, eta, eta1, G11a, G12a, G21a, G22a)
     call G_ab_RPT(kxmu, eta1, eta0, G11c, G12c, G21c, G22c)
  elseif(itype == 1) then
     call G_ab_Closure(k, eta, eta1, G11a, G12a, G21a, G22a)
     call G_ab_Closure(kxmu, eta1, eta0, G11c, G12c, G21c, G22c)
  endif

  gamma112 = 0.5d0 * (1.d0 - x * mu) / xabs2
  gamma121 = 0.5d0 * mu / x
  gamma222 = 0.5d0 * ( mu - x ) / xabs2 / x

  call Lambda(itype, k*x, y, nu, yabs2, eta1, eta0, I1, I2)

  A = gamma112 * I1 * ( G21c + G22c ) + gamma121 * I2 * ( G11c + G12c )
  B = gamma222 * I2 * ( G21c + G22c )

  int_J1 = G11a * A + G12a * B
  int_J2 = G21a * A + G22a * B

  return
end subroutine int_J
! ******************************************************* !
!
subroutine get_func_K(itype, k, x, y, mu, nu, xabs2, yabs2, &
     & eta, eta0, func_K1, func_K2)
!
! ******************************************************* !
!
!     computing function K_1 & K2 using method of Gaussian quadratures 
!
  implicit none
  integer, parameter :: ieta_max = 40
  integer, intent(in) :: itype
  double precision, intent(in) :: k, x, y, mu, nu, xabs2, yabs2, eta, eta0
  double precision, intent(out) :: func_K1, func_K2
  integer iitype, ieta
  double precision int_K1, int_K2, func_K1a
  double precision k1, k2, k3, k4, k5, ss
  double precision ff, gg, hh, ii, jj
  double precision weta(ieta_max), xeta(ieta_max)
  double precision fact, xy
  double precision ef, eg, eh, ei, ej
! ------------------------------------------

  k1 = k
  k2 = k * x
  k3 = k * sqrt(xabs2)
  k4 = k3 * y
  k5 = k3 * sqrt(yabs2)

  if(max(k1,k2,k3,k4,k5) <= 5d-3) then 
!     using the linear propagator, integration is analytically performed
     ss = eta - eta0

     xy = x * y
     fact = xabs2 * yabs2 * xy

     eh = exp(ss)
     eg = eh * eh         ! exp(2ss)    
     ef = eg * eh         ! exp(3ss)
     ei = 1.d0 / sqrt(eh) ! exp(-0.5ss)
     ej = ei / eh         ! exp(-1.5ss)

     ! //// function K1 ///// ! 
     ff = -28.d0*x*x*y*mu * ( nu*nu - 1.d0 )            &
          & + mu * ( 19.d0*y + 63.d0*nu - 82.d0*y*nu*nu ) &
          & + 7.d0 *x* ( 5.d0 - 14.d0*mu*mu ) * nu        &
          & + x*y * ( -5.d0 - 30.d0*nu*nu                 &
          & + 14.d0*mu*mu * ( 10.d0*nu*nu - 3.d0 ) )
     ff = ff / 504.d0
     gg = ( -7.d0*mu + x*(-3.d0 + 10.d0*mu*mu ) )   &
          & * ( -5.d0*nu + y*(-1.d0 + 6.d0*nu*nu ) ) 
     gg = -gg / 140.d0
     hh = 4.d0*x*x*y*mu * ( nu*nu - 1.d0 )      &
          & + mu * ( y + 5.d0*nu - 6.d0*y*nu*nu ) &
          & + x * ( 1.d0 - 6.d0*mu*mu ) * nu      &
          & + x*y * ( 1.d0 - 2.d0*nu*nu           & 
          & + 2.d0*mu*mu * ( 2.d0*nu*nu + 1.d0 ) )
     hh = hh / 40.d0
     ii = -2.d0/35.d0 * ( x - mu ) * ( nu*nu - 1.d0 ) * y
     jj = 7.d0 * y * mu * ( nu*nu - 1.d0 ) & 
          & + 7.d0*x*x*y*mu * ( nu*nu - 1.d0 )  &
          & - 2.d0*x* ( 1.d0 - mu*mu ) * nu                            &
          & - 2.d0*x*y * ( -4.d0 + 3.d0*nu*nu + mu*mu * ( -3.d0 + 4.d0*nu*nu ) )
     jj = -2.d0/315.d0 * jj

     func_K1 = ff * ef + gg * eg + hh * eh + ii * ei + jj * ej
     func_K1 = func_K1 / fact

     ! //// function K2 ///// !
     ff = y * ( x + mu + 4.d0*x*x*mu - 6.d0*x*mu*mu )  &  
          & - 7.d0*nu * ( x - 3.d0*mu + 2.d0*x*mu*mu ) &
          & + 2.d0*y*nu*nu * ( -11.d0*mu + x * (3.d0 - 2.d0*x*mu + 10.d0*mu*mu) )  
     ff = ff / 168.d0
     gg = ( -7.d0*mu + x*( 1.d0 + 6.d0*mu*mu ) )    & 
          & * ( -5.d0*nu + y*(-1.d0 + 6.d0*nu*nu ) ) 
     gg = -gg / 140.d0
     ii = ( -3.d0*mu + x * (4.d0 - 5.d0*x*mu + 4.d0*mu*mu) ) &
          & * ( nu*nu - 1.d0 ) * y / 35.d0
     jj = -jj * 1.5d0 

     func_K2 = ff * ef + gg * eg + hh * eh + ii * ei + jj * ej
     func_K2 = func_K2 / fact

  elseif(max(k1,k2,k3,k4,k5) > 5d-3) then 
!     numerical integration by Gaussian quadratures
     func_K1 = 0.d0
     func_K2 = 0.d0

     call gauleg(eta0, eta, xeta, weta, ieta_max)
     
     do ieta = 1, ieta_max
        call int_K(k, x, y, mu, nu, xabs2, yabs2, eta, eta0, xeta(ieta), &
             & itype, int_K1, int_K2)
        func_K1 = func_K1 + weta(ieta) * int_K1
        func_K2 = func_K2 + weta(ieta) * int_K2
     enddo

  endif
  return
end subroutine get_func_K
! ******************************************************* !
!
subroutine int_K(k, x, y, mu, nu, xabs2, yabs2, eta, eta0, eta1, &
     & itype, int_K1, int_K2)
!
! ******************************************************* !

  implicit none
  integer, intent(in) :: itype
  double precision, intent(in) :: k, x, y, mu, nu, xabs2, yabs2, eta, eta0, eta1
  double precision, intent(out) ::  int_K1, int_K2
  double precision G11a, G12a, G21a, G22a, G11c, G12c, G21c, G22c
  double precision I1, I2, A, B
  double precision gamma112, gamma121, gamma222
! ---------------------------------

  if(itype == 0) then 
     call G_ab_RPT(k, eta, eta1, G11a, G12a, G21a, G22a)
     call G_ab_RPT(k*x, eta1, eta0, G11c, G12c, G21c, G22c)
  elseif(itype == 1) then
     call G_ab_Closure(k, eta, eta1, G11a, G12a, G21a, G22a)
     call G_ab_Closure(k*x, eta1, eta0, G11c, G12c, G21c, G22c)
  endif

  gamma112 = 0.5d0 * (1.d0 - x * mu) / xabs2 
  gamma121 = 0.5d0 * mu / x
  gamma222 = 0.5d0 * ( mu - x ) / xabs2 / x

  call Lambda(itype, k*sqrt(xabs2), y, nu, yabs2, eta1, eta0, I1, I2)

  A = gamma112 * ( G11c + G12c ) * I2 + gamma121 * ( G21c + G22c ) * I1
  B = gamma222 * ( G21c + G22c ) * I2

  int_K1 = G11a * A + G12a * B 
  int_K2 = G21a * A + G22a * B 
 
  return
end subroutine int_K
!
! ******************************************************* !
!
subroutine Lambda(itype, k, x, mu, xabs2, eta1, eta0, Lambda1, Lambda2)
!
! ******************************************************* !
!     computing the function Lambda by Gaussian quadratures
!
!     itype : RPT or  CLA
!
  implicit none
  
  integer, parameter :: ieta_max = 40
  integer, intent(in) :: itype
  double precision, intent(in) :: k, x, mu, xabs2, eta1, eta0
  double precision, intent(out) :: Lambda1, Lambda2
  integer ieta
  double precision k1, k2, k3
  double precision int_lambda1, int_lambda2, ss
  double precision weta(ieta_max), xeta(ieta_max)
  double precision mu2, xmu2, A1, A2, A3, B1, fact
! ---------------------------------------

  k1 = k
  k2 = k * x
  k3 = k * sqrt(xabs2)

  if(max(k1,k2,k3) <= 5d-3) then 
!
!     using the linear propagator, integration is analytically performed
!
     ss = eta1 - eta0 

     Lambda1 = 4.d0 * x * (-1.d0 + mu*mu) * exp(-1.5d0*ss) / 35.d0 &
          & - 0.1d0 * (x + 5.d0*mu -6.d0*x*mu*mu) * exp(ss) &
          & + (7.d0*mu + x* (3.d0  - 10.d0*mu*mu) ) * exp(2.d0*ss) / 14.d0
     Lambda1 = Lambda1 / x / xabs2
     
     Lambda2 = -6.d0 * x * (-1.d0 + mu*mu) * exp(-1.5d0*ss) / 35.d0 &
          & - 0.1d0 * (x + 5.d0*mu -6.d0*x*mu*mu) * exp(ss) &
          & + (7.d0*mu - x - 6.d0*x*mu*mu ) * exp(2.d0) / 14.d0
     Lambda2 = Lambda2 / x / xabs2
 
  elseif(max(k1,k2,k3) > 5d-3) then 
!     integral over "eta" by Gaussian quadratures 
     Lambda1 = 0.d0
     Lambda2 = 0.d0
     call gauleg(eta0, eta1, xeta, weta, ieta_max)

     do ieta = 1, ieta_max
        call int_lambda(k, x, mu, xabs2, eta0, eta1, xeta(ieta), &
             & itype, int_lambda1, int_lambda2)
        Lambda1 = Lambda1 + weta(ieta) * int_lambda1
        Lambda2 = Lambda2 + weta(ieta) * int_lambda2
     enddo
  endif
  return
end subroutine Lambda
!
! ******************************************************* !
!
subroutine int_lambda(k, x, mu, xabs2, eta0, eta1, eta2, &
     & itype, int_lambda1, int_lambda2)
!
! ******************************************************* !
!
!     integrand of function Lambda, as function of eta_s
!
  implicit none
  integer, intent(in) :: itype
  double precision, intent(in) :: k, x, mu, xabs2, eta0, eta1, eta2
  double precision, intent(out) :: int_lambda1, int_lambda2
  double precision k1, k2, k3, A, B
  double precision G11a, G12a, G21a, G22a
  double precision G11b, G12b, G21b, G22b
  double precision G11c, G12c, G21c, G22c
  double precision gamma112, gamma121, gamma222
! ------------------------------------------

  k1 = k
  k2 = k * x
  k3 = k * sqrt(xabs2)

  if(itype == 0) then 
     call G_ab_RPT(k1, eta1, eta2, G11a, G12a, G21a, G22a)
     call G_ab_RPT(k2, eta2, eta0, G11b, G12b, G21b, G22b)
     call G_ab_RPT(k3, eta2, eta0, G11c, G12c, G21c, G22c)
  elseif(itype == 1) then
     call G_ab_Closure(k1, eta1, eta2, G11a, G12a, G21a, G22a)
     call G_ab_Closure(k2, eta2, eta0, G11b, G12b, G21b, G22b)
     call G_ab_Closure(k3, eta2, eta0, G11c, G12c, G21c, G22c)
  endif

  gamma112 = 0.5d0 * (1.d0 - x * mu) / xabs2
  gamma121 = 0.5d0 * mu / x
  gamma222 = 0.5d0 * ( mu - x ) / xabs2 / x

  A = gamma112 * (G11b + G12b)*(G21c + G22c) &
       & + gamma121 * (G21b + G22b) * (G11c + G12c)
  B = gamma222 * (G21b + G22b) * (G21c + G22c)

  int_lambda1 = G11a * A + G12a * B
  int_lambda2 = G21a * A + G22a * B 

  return
end subroutine int_lambda

! ************************************************************

subroutine gauleg(x1,x2,x,w,n)

! ************************************************************
  
  double precision, parameter :: EPS = 3d-14
  double precision, parameter :: pi = 3.14159265358979323846d0
  integer, intent(in) :: n
  double precision, intent(in) :: x1, x2
  double precision, intent(out) :: x(n), w(n)
  integer i, j, m
  double precision p1, p2, p3, pp, xl, xm, z, z1

  m = (n+1)/2
  xm = 0.5d0*(x2+x1)
  xl = 0.5d0*(x2-x1)
  do i = 1, m
     z = cos(pi*(dble(i)-0.25d0)/(dble(n)+0.5d0))
1    continue
     p1 = 1.d0
     p2 = 0.d0
     do j = 1, n
        p3 = p2
        p2 = p1
        p1 = ((2.d0*dble(j)-1.d0)*z*p2-(dble(j)-1.d0)*p3)/dble(j)
     enddo
     pp = dble(n)*(z*p1-p2)/(z*z-1.d0)
     z1 = z
     z = z1-p1/pp
     if(abs(z-z1) > EPS)goto 1
     x(i) = xm - xl*z
     x(n+1-i) = xm + xl*z
     w(i) = 2.d0*xl/((1.d0-z*z)*pp*pp)
     w(n+1-i) = w(i)
  enddo
  return
end subroutine gauleg
!  (C) Copr. 1986-92 Numerical Recipes Software 

